/**@file
   @brief Windowsの WndProc を横取りするための仕掛け。実装部。

   @attention 2013-03-11

   Windowをサブクラス化する際、サブクラス化されるWindowがUnicodeバージョンで作ら
   れているかAnsiバージョンで作られているかによって、CallWindowProcで自動的な文字
   コード変換が行なわれるようになっている。何故このレベルで文字コード変換が必要な
   のか当初イマイチ理解出来なかったのだけれども、事実そうなっている。

   なぜか。

   例えば、SetWindowTextWを使ってウィンドウタイトルをセットしようとした場合、
   WM_SETTEXTメッセージのLPARAMにはutf-16le（wchar_t）型の文字列が指定される。
   SetWindowTextAを使ってウィンドウタイトルをセットする場合はchar文字列である。こ
   の場合、どちらの文字列を使っているかを示すのはメッセージ番号ではなく、どのAPI
   を使って送られたかである。SendMessageA / PostMessageA で送られたのであれば
   char文字列、SendMessageW / PostMessageW で送られたのであればutf-16le文字列と判
   定される。間違いない。

   また、メッセージを処理する側は、UNICODE対応アプリケーションであれば、
   GetMessageW / PeekMessageW を、そうでなければ GetMessageA / PeekMessageA を使っ
   ている。そこで、パラメータに文字列を含むメッセージがSendMessageAで送
   られ、それを受け取る側がGetMessageWを使っている場合、間に入ったWindowsは文字コー
   ドの変換を行なうようになっているのである。オーバーヘッドを考えれば、たぶん、読
   み出す側で。

   更に、メッセージのディスパッチに使われるAPIも違う。UNICODEアプリなら
   DispatchMessageWが、AnsiアプリならDispatchMessageAが使われているはずだけれども、
   サブクラス化する際にはこの段階で横取りが行なわれている筈である。つまり、呼び出
   されるウィンドウプロシージャにも、UNICODE対応版とANSI版の二種類があるというこ
   とである。ここまでのメカニズムを考えれば当たり前の話だが、WNDPROCのインター
   フェース的には違いは無いのでプロシージャ自体は共用出来る。

   ところが、Unicodeで作られているウィンドウに対してSetWindowLongAでWNDPROCを登録
   すると、登録されたWNDPROCはAnsi版だと判定されて、GetMessageWによってutf-16leに
   なっている文字列はAnsi版向けのコード変換が行なわれた状態で呼び出されることにな
   る。そこからCallWindowProcAで元のウィンドウプロシージャを呼び出すと、（元の
   WNDPROCはUnicde版なので）再度Unicodeへの変換が行なわれるという流れになる。この
   時、変換出来ない文字は化ける。

   同じケースの場合、SetWindowLongW でWNDPROCを登録し、CallWindowProcWで元のウィ
   ンドウプロシージャを呼び出すようにすると、A系のAPIで呼び出された時に
   GetMessageWのところで変換が発生するまでは同じだけれども、以後コード変換は発生
   しない。

   Ansi版のウィンドウをサブクラス化する場合はコード変換が出来ない文字があることは
   外部化されているので（つまりそれは仕様）、文字が化けても問題は無い。しかしなが
   ら、無駄な変換処理を発生させることはいずれにしても好ましくないので、サブクラス
   化する対象がどちらのタイプのメッセージループを持っているかを判定した上で、
   SetWindowLongA/W と CallWindowProcA/W をペアにして使うのが一番いい。（今どきの
   アプリであれば、W系決め打ちでも実質的に問題は発生しないのかも知れない。少なく
   ともWindowsのComponentは全部Unicode版で作られているはず）

   やれやれ。

**/

#include "ALWndProcHook.hpp"
#include "ALUtSystem.hpp"

#include <WSCbase.h>
#include <WSDdev.h>

#include <WSCbase.h>

namespace al {
namespace system {


std::multimap<HWND,CALWndProcHook::WPReplacement>	CALWndProcHook::_wp_replacement_map;

std::map<HWND,CALWndProcHook::WndProc_t*>	CALWndProcHook::_old_wndproc_map;

CALWndProcHook::~CALWndProcHook()
{
    UnregisterHook();
}

CALWndProcHook::WndProc_t* CALWndProcHook::ReplaceWndProc(
    WSCbase* win,
    WndProc_t* new_wndproc )
{
    HWND hwnd = reinterpret_cast<HWND>(GetWindowHWND(win));
    if( hwnd == NULL ) return NULL;

    WndProc_t* old_proc;
    if( IsWindowUnicode(hwnd) )
        old_proc = (WNDPROC)SetWindowLongW(hwnd,GWL_WNDPROC,(DWORD)new_wndproc);
    else
        old_proc = (WNDPROC)SetWindowLongA(hwnd,GWL_WNDPROC,(DWORD)new_wndproc);

    return old_proc;
}

HWND CALWndProcHook::RegisterHook(WSCbase* win, bool use_special)
{
    HWND hwnd;
    if( use_special )
        hwnd = reinterpret_cast<HWND>(GetSpecialHWND(win));
    else
        hwnd = reinterpret_cast<HWND>(GetWindowHWND(win));
    if( hwnd == NULL ) return NULL;

    /** WndProc の置換は一つの HWND に対して一回だけ */
    if( _old_wndproc_map.find(hwnd) == _old_wndproc_map.end() ) {
        WndProc_t* old_proc;
        if( IsWindowUnicode(hwnd) )
            old_proc = (WndProc_t*)SetWindowLongW(hwnd,GWL_WNDPROC,(DWORD)RawWndProc);
        else
            old_proc = (WndProc_t*)SetWindowLongA(hwnd,GWL_WNDPROC,(DWORD)RawWndProc);
        _old_wndproc_map.insert(std::pair<HWND,WndProc_t*>(hwnd,old_proc));
    }

    /** ここから呼び出される WndProc はクラスのインスタンス単位で
        複数登録出来る。 */
    WPReplacement	replacement(win,this);
    _wp_replacement_map.insert(std::pair<HWND,WPReplacement>(hwnd,replacement));
    return hwnd;
}

void CALWndProcHook::UnregisterHook()
{
    HWND hwnd = NULL;

    /** このインスタンスが登録されている _wp_replacement_map 要素を削除する。<br>
        どのHWNDに対して登録されているかも同時に調べる。 */

    std::multimap<HWND,WPReplacement>::iterator	wp_it;
    for( wp_it = _wp_replacement_map.begin();
         wp_it != _wp_replacement_map.end(); ) {
        if( wp_it->second._pthis == this ) {
            hwnd = wp_it->first;
            _wp_replacement_map.erase(wp_it++);
        }
        else
            ++wp_it;
    }

    /** 登録されているHWNDが無くなったら WndProc を元に戻す。 */

    if( hwnd != NULL
        && _wp_replacement_map.find(hwnd) == _wp_replacement_map.end() ) {
        std::map<HWND,WndProc_t*>::iterator	ow_it = _old_wndproc_map.find(hwnd);
        if( ow_it != _old_wndproc_map.end() ) {
            if( IsWindowUnicode(hwnd) )
                SetWindowLongW(hwnd,GWL_WNDPROC,(DWORD)ow_it->second);
            else
                SetWindowLongA(hwnd,GWL_WNDPROC,(DWORD)ow_it->second);
            _old_wndproc_map.erase(ow_it);
        }
    }
}

LRESULT CALLBACK CALWndProcHook::RawWndProc(
    HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam
    )
{
    typedef std::multimap<HWND,WPReplacement>::iterator	iter;
    std::pair<iter,iter>	wp_range = _wp_replacement_map.equal_range(hWnd);

    LRESULT	lResult=0;

    if( message == WM_CLOSE ) {
        lResult=1;
    }

    iter	it;
    for( it = wp_range.first; it != wp_range.second; ++it ) {
        if( it->second._pthis->MyWndProc(lResult,it->second._win,
                                         hWnd,message,wParam,lParam) )
            return lResult;
    }

    std::map<HWND,WndProc_t*>::iterator	ow_it = _old_wndproc_map.find(hWnd);
    if( ow_it != _old_wndproc_map.end() ) {
        if( IsWindowUnicode(hWnd) )
            lResult = CallWindowProcW(ow_it->second,hWnd,message,wParam,lParam);
        else
            lResult = CallWindowProcA(ow_it->second,hWnd,message,wParam,lParam);
    }

    return lResult;
}


} /* namespace system */
} /* namespace al */
