// -*- tab-width:4 -*-
#pragma once
#if !defined(ALWNDPROCHOOK_HPP_)
#define ALWNDPROCHOOK_HPP_

/**@file
   @brief Windowsの WndProc を横取りするための仕掛け。
**/

#include <Windows.h>
#include <map>
#include "ALException.hpp"

class WSCbase;

namespace al {
namespace system {

/**
   WndProc置換と派生クラスのメンバ関数へのコールバックを行なうための基底クラス。
 **/
class CALWndProcHook {
public:
    typedef LRESULT (CALLBACK WndProc_t)(
        HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam
        );

                                /** WndProc置換処理。単体で汎用的に利用可能。 */
    static WndProc_t* ReplaceWndProc(WSCbase* win, WndProc_t* new_wndproc);

protected:
    CALWndProcHook() {}
    explicit CALWndProcHook(WSCbase* win,bool use_special=false)
    {
        if( !RegisterHook(win,use_special) )
            ::al::ThrowALException(::al::ALException());
    }

    virtual ~CALWndProcHook();

    /** win に対するウィンドウプロシージャ処理インスタンスとして自分自身を登録する。<br>
        use_special == true の時は GetSpecialHWND() で返されるハンドルを登録する。
        use_special == false の時は GetWindowHWND()で返されるハンドルを登録する。
        @return 登録のキーとなった HWND。登録出来なかった時はNULLを返す。*/
    HWND RegisterHook(WSCbase* win,bool use_special=false);

    /** インスタンスの登録を解除する。 */
    void UnregisterHook();

    /** RawWndProcから呼び出されるクラスのウィンドウプロシージャ。
        戻り値はメッセージを処理した場合（＝後続のWndProc処理が不要な場合）true。
        処理結果のコードは lResult に格納して返すこと。 */
    virtual bool MyWndProc(
        LRESULT& lResult,
        WSCbase* win,
        HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam
        ) = 0;

private:

                                /** 直接ウィンドウメッセージを受け取るハンドラ。
                                    ここでHWNDの値を元に交通整理する。 */
    static LRESULT CALLBACK RawWndProc(
        HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam
        );

                                /** 交通整理するための情報を格納する構造体 */
    struct WPReplacement {
        WSCbase*	_win;
        CALWndProcHook*	_pthis;
        WPReplacement() : _win(NULL), _pthis(NULL) {}
        WPReplacement(WSCbase* win,CALWndProcHook* pthis)
            : _win(win), _pthis(pthis) {}
    };

                                /** 交通整理するための情報のmap */
    static std::multimap<HWND,WPReplacement>	_wp_replacement_map;

                                /** このクラスが介入する前に登録されていた
                                    オリジナルのWndProcへのポインタのマップ。 */
    static std::map<HWND,WndProc_t*>	_old_wndproc_map;

};

} /* namespace system */
} /* namespace al */



#endif /*ALWNDPROCHOOK_HPP_*/
